/**Lichtempfaenger V3
   Controls up to 10 lights.

  Uses: 
   ESPNOW - Basic communication - Slave
   Date: 26th September 2017
   Author: Arvind Ravulavaru <https://github.com/arvindr21>
   Purpose: ESPNow Communication between a Master ESP32 and multiple ESP32 Slaves
   Description: This sketch consists of the code for the Slave module.
   Resources: (A bit outdated)
   a. https://espressif.com/sites/default/files/documentation/esp-now_user_guide_en.pdf
   b. http://www.esploradores.com/practica-6-conexion-esp-now/

   << This Device Slave >>

   Flow: Master
   Step 1 : ESPNow Init on Master and set it in STA mode
   Step 2 : Start scanning for Slave ESP32 (we have added a prefix of `slave` to the SSID of slave for an easy setup)
   Step 3 : Once found, add Slave as peer
   Step 4 : Register for send callback
   Step 5 : Start Transmitting data from Master to Slave(s)

   Flow: Slave
   Step 1 : ESPNow Init on Slave
   Step 2 : Update the SSID of Slave with a prefix of `slave`
   Step 3 : Set Slave in AP mode
   Step 4 : Register for receive callback and wait for data
   Step 5 : Once data arrives, print it in the serial monitor

   Note: Master and Slave have been defined to easily understand the setup.
         Based on the ESPNOW API, there is no concept of Master and Slave.
         Any devices can act as master or salve.
*/


#define COM_ESPNOW 0
#define COM_BLE 1

#define COM COM_ESPNOW


#if COM == COM_ESPNOW
  #ifdef ESP32
    #include <esp_now.h>
  #else
    #include <espnow.h>
  #endif
#endif
#if COM == COM_BLE
  #include <BLEDevice.h>
#endif


#ifdef ESP32
  #include <WiFi.h>
  #define USE_PINS // control the HW pins listed in lpin
#else
  #include <ESP8266WiFi.h>
  #include <ESP8266WebServer.h>
  #define WebServer ESP8266WebServer
#endif

#include <Adafruit_NeoPixel.h>
#include <ESP2SOTA.h>
#include <Preferences.h>


// global config
#define VERSION "3.1"

#define CHANNEL 1 // WIFI channel

// Which pin on the Arduino is connected to the NeoPixels?
#ifdef ESP32
#define PIN       16
#else
#define PIN       15 // D8
#endif

// How many NeoPixels are attached to the Arduino?
#define NUMPIXELS 10
#define LED_COLOR_R_ON 200
#define LED_COLOR_G_ON 100
#define LED_COLOR_B_ON 30
#define LED_COLOR_R_OFF 0
#define LED_COLOR_G_OFF 0
#define LED_COLOR_B_OFF 0

// Wifi config
#define WIFI_PASSWORD "esp32licht"
#define MAX_NUM_SLAVES 10

// Licht pins
const byte lpin[NUMPIXELS] = {
  27,
  25,
  32,
  12,
  4,
  22,
  21,
  17,
  23,
  19,
};

enum lmodes_et {
  MODE_NORMAL_E,
  MODE_NEON_E,
  MODE_DEFECT_E,
  MODE_FIRE_E,
  MODE_TV_E
  //etc
};
typedef struct licht_st
{
  unsigned short l[MAX_NUM_SLAVES]; // all slaves receive all data, depending on slave number will pick an element
} licht_s;
licht_s myData;
byte lcolor[NUMPIXELS][3], lmode[NUMPIXELS], lold[NUMPIXELS];
int lstate[NUMPIXELS];
uint32_t ltime[NUMPIXELS];

String mainForm PROGMEM = R"rawliteral(
<!DOCTYPE HTML><html><head>
<title>Lichtempf&auml;nger</title>
<meta name="viewport" content="width=device-width, initial-scale=1">
</head><body>
<h3>Konfiguration Lichtempf&auml;nger</h3>
<p>Version: #VERSION<br>
Slave-ID: #SLAVEID<br>
<a href=/prepupdate>Update Firmware</a><br>
<br></p>
<form action="/get">
Slave-ID: <input type="number" min="0" max="9" size=3 name="slaveid" placeholder="0-9" value="#SLAVEID"> Wichtig: Eine &Auml;nderung hier erfordert eine neue WLAN-Verbindung!<br>
<br>
Licht 1: <a href=#CL1X</a> <input type="checkbox" id="cbl1" name="l1" #L1C> Farbe (RGB) <input type="number" min="0" max="255" size=4 name="l1r" placeholder="Rot 0-255" value="#L1R"> 
         <input type="number" min="0" max="255" size=4 name="l1g" placeholder="Gruen 0-255" value="#L1G"> 
         <input type="number" min="0" max="255" size=4 name="l1b" placeholder="Blau 0-255" value="#L1B">,
         Modus (0-4) <input type="number" min="0" max="4" size=3 name="l1m" placeholder="0=normal, 1=neon, 2=kaputt, 3=Feuer, 4=TV" value="#L1M"><br>
<br>
Licht 2: <a href=#CL2X</a> <input type="checkbox" id="cbl2" name="l2" #L2C> Farbe (RGB) <input type="number" min="0" max="255" size=4 name="l2r" placeholder="Rot 0-255" value="#L2R"> 
         <input type="number" min="0" max="255" size=4 name="l2g" placeholder="Gruen 0-255" value="#L2G"> 
         <input type="number" min="0" max="255" size=4 name="l2b" placeholder="Blau 0-255" value="#L2B">,
         Modus (0-4) <input type="number" min="0" max="4" size=3 name="l2m" placeholder="0=normal, 1=neon, 2=kaputt, 3=Feuer, 4=TV" value="#L2M"><br>
<br>
Licht 3: <a href=#CL3X</a> <input type="checkbox" id="cbl3" name="l3" #L3C> Farbe (RGB) <input type="number" min="0" max="255" size=4 name="l3r" placeholder="Rot 0-255" value="#L3R"> 
         <input type="number" min="0" max="255" size=4 name="l3g" placeholder="Gruen 0-255" value="#L3G"> 
         <input type="number" min="0" max="255" size=4 name="l3b" placeholder="Blau 0-255" value="#L3B">,
         Modus (0-4) <input type="number" min="0" max="4" size=3 name="l3m" placeholder="0=normal, 1=neon, 2=kaputt, 3=Feuer, 4=TV" value="#L3M"><br>
<br>
Licht 4: <a href=#CL4X</a> <input type="checkbox" id="cbl4" name="l4" #L4C> Farbe (RGB) <input type="number" min="0" max="255" size=4 name="l4r" placeholder="Rot 0-255" value="#L4R"> 
         <input type="number" min="0" max="255" size=4 name="l4g" placeholder="Gruen 0-255" value="#L4G"> 
         <input type="number" min="0" max="255" size=4 name="l4b" placeholder="Blau 0-255" value="#L4B">,
         Modus (0-4) <input type="number" min="0" max="4" size=3 name="l4m" placeholder="0=normal, 1=neon, 2=kaputt, 3=Feuer, 4=TV" value="#L4M"><br>
<br>
Licht 5: <a href=#CL5X</a> <input type="checkbox" id="cbl5" name="l5" #L5C> Farbe (RGB) <input type="number" min="0" max="255" size=4 name="l5r" placeholder="Rot 0-255" value="#L5R"> 
         <input type="number" min="0" max="255" size=4 name="l5g" placeholder="Gruen 0-255" value="#L5G"> 
         <input type="number" min="0" max="255" size=4 name="l5b" placeholder="Blau 0-255" value="#L5B">,
         Modus (0-4) <input type="number" min="0" max="4" size=3 name="l5m" placeholder="0=normal, 1=neon, 2=kaputt, 3=Feuer, 4=TV" value="#L5M"><br>
<br>
Licht 6: <a href=#CL6X</a> <input type="checkbox" id="cbl6" name="l6" #L6C> Farbe (RGB) <input type="number" min="0" max="255" size=4 name="l6r" placeholder="Rot 0-255" value="#L6R"> 
         <input type="number" min="0" max="255" size=4 name="l6g" placeholder="Gruen 0-255" value="#L6G"> 
         <input type="number" min="0" max="255" size=4 name="l6b" placeholder="Blau 0-255" value="#L6B">,
         Modus (0-4) <input type="number" min="0" max="4" size=3 name="l6m" placeholder="0=normal, 1=neon, 2=kaputt, 3=Feuer, 4=TV" value="#L6M"><br>
<br>
Licht 7: <a href=#CL7X</a> <input type="checkbox" id="cbl7" name="l7" #L7C> Farbe (RGB) <input type="number" min="0" max="255" size=4 name="l7r" placeholder="Rot 0-255" value="#L7R"> 
         <input type="number" min="0" max="255" size=4 name="l7g" placeholder="Gruen 0-255" value="#L7G"> 
         <input type="number" min="0" max="255" size=4 name="l7b" placeholder="Blau 0-255" value="#L7B">,
         Modus (0-4) <input type="number" min="0" max="4" size=3 name="l7m" placeholder="0=normal, 1=neon, 2=kaputt, 3=Feuer, 4=TV" value="#L7M"><br>
<br>
Licht 8: <a href=#CL8X</a> <input type="checkbox" id="cbl8" name="l8" #L8C> Farbe (RGB) <input type="number" min="0" max="255" size=4 name="l8r" placeholder="Rot 0-255" value="#L8R"> 
         <input type="number" min="0" max="255" size=4 name="l8g" placeholder="Gruen 0-255" value="#L8G"> 
         <input type="number" min="0" max="255" size=4 name="l8b" placeholder="Blau 0-255" value="#L8B">,
         Modus (0-4) <input type="number" min="0" max="4" size=3 name="l8m" placeholder="0=normal, 1=neon, 2=kaputt, 3=Feuer, 4=TV" value="#L8M"><br>
<br>
Licht 9: <a href=#CL9X</a> <input type="checkbox" id="cbl9" name="l9" #L9C> Farbe (RGB) <input type="number" min="0" max="255" size=4 name="l9r" placeholder="Rot 0-255" value="#L9R"> 
         <input type="number" min="0" max="255" size=4 name="l9g" placeholder="Gruen 0-255" value="#L9G"> 
         <input type="number" min="0" max="255" size=4 name="l9b" placeholder="Blau 0-255" value="#L9B">,
         Modus (0-4) <input type="number" min="0" max="4" size=3 name="l9m" placeholder="0=normal, 1=neon, 2=kaputt, 3=Feuer, 4=TV" value="#L9M"><br>
<br>
Licht 10: <a href=#CL10X</a> <input type="checkbox" id="cbl10" name="l10" #L10C> Farbe (RGB) <input type="number" min="0" max="255" size=4 name="l10r" placeholder="Rot 0-255" value="#L10R"> 
         <input type="number" min="0" max="255" size=4 name="l10g" placeholder="Gruen 0-255" value="#L10G"> 
         <input type="number" min="0" max="255" size=4 name="l10b" placeholder="Blau 0-255" value="#L10B">,
         Modus (0-4) <input type="number" min="0" max="4" size=3 name="l10m" placeholder="0=normal, 1=neon, 2=kaputt, 3=Feuer, 4=TV" value="#L10M"><br>
<br>Verf&uuml;gbare Modi (Effekte):<br>
<ul><li>0=normal</li>
<li>1=Neonlampe</li>
<li>2=kaputte Neonlampe (flackert immer)</li>
<li>3=Feuer*</li>
<li>4=Fernseher*</li></ul>
* funktioniert nicht mit externen LEDs an den Ausgängen.
<br>
<br><input type="submit" value="&Auml;ndern (und Neustarten bei Slave-ID-Wechsel)">
</form></body></html>)rawliteral";
String prepupdateForm PROGMEM = 
  "<!DOCTYPE HTML><html><head><title>Lichtempf&auml;nger</title><meta name=\"viewport\" content=\"width=device-width, initial-scale=1\">"
  "</head><body><h3>Firmwareupdate Lichtempf&auml;nger</h3><p>"
  "Bitte bereite eine BIN-Datei vor.<br>In Arduino Sketch->Export Compiled Binary ausw&auml;hlen.<br>"
  "<br><a href=/update>Zum Update</a><br><br><a href=/>Zur&uuml;ck</a></p></body></html>";
String reboot PROGMEM =
  "<!DOCTYPE HTML><html><head><title>Lichtempf&auml;nger</title><meta name=\"viewport\" content=\"width=device-width, initial-scale=1\"></head>"
  "<body>Neustart, bitte warten...<br><a href=/>Zur&uuml;ck</a></body></html>";
String noreboot PROGMEM =
  "<!DOCTYPE HTML><html><head><title>Lichtempf&auml;nger</title><meta name=\"viewport\" content=\"width=device-width, initial-scale=1\"><meta http-equiv=\"Refresh\" content=\"3; url=/\" /></head>"
  "<body>Einstellungen gespeichert.<br><a href=/>Zur&uuml;ck</a></body></html>";

WebServer server(80);
bool updatemode;
Preferences preferences;
byte SLAVE_ID;
Adafruit_NeoPixel pixels(NUMPIXELS, PIN, NEO_GRB + NEO_KHZ800);
bool toggle, rx;


// Neon from https://forum.arduino.cc/t/leuchtstoffrohren-simulieren/693743/10
//                aus, zündet, leuchtet, 100%
enum class State {OFF, START, RUNNING, FULL};
#define NEON_MAX 255
#define NEON_ON 100
#define NEON_STEP 1
#define NEON_WARMUP_INTERVAL 400
class NeonLed {
  private:
    unsigned long previousMillis;        // timestamp for states
    unsigned long previousMillisEffect;  // timestamp for the effect
    uint16_t interval  = 10;             // how fast should the led be dimmed (=milliseconds between steps)
    uint16_t startTimeMin = 3000;        // how long will it take from off to stable on
    uint16_t startTimeMax = 6000;

  public:
    State state = State::OFF;        
    byte actual = 0;                     // actual PWM
    NeonLed() {}

    void on()        // einschalten
    {
      state = State::START;
      previousMillis = millis();
      previousMillisEffect = previousMillis;
      actual = 2;
      interval = random(200, 800);
      //Serial.print("on "); Serial.println(millis());
    }

    void off()       // ausschalten
    {
      actual = 0;
      //Serial.print("off "); Serial.println(millis());
      state = State::OFF;
    }

    void tick()
    {
      if (state == State::START)
      {
        uint32_t currentMillis = millis();
        if (currentMillis - previousMillisEffect > interval)
        {
          if (actual >= NEON_ON)           // alles ab diesem Wert ist "ein" daher müssen wir nun "aus" Schalten
          {
            actual = random(1, 3);         // leichtes Glimmen der "Enden" ... könnte man auch ganz auf 0 setzen
            interval = random(700, 2000);  // unregeläßige Dunkelphasen (zwischen dem Aufblitzen)
            //Serial.print("gl "); Serial.println(millis());
          }
          else
          {
            actual = NEON_ON; //random(NEON_ON, 255);      
            interval = random(30, 60);                 // kurzes Aufblitzen
            //Serial.print("fl "); Serial.println(millis());
          }
          previousMillisEffect = currentMillis;
        }
        if (currentMillis - previousMillis > (uint16_t)random(startTimeMin, startTimeMax))
        {
          actual = NEON_ON;      // after the neon is stable "on" it will take some time to reach 100%
          //Serial.print("st "); Serial.println(millis());
          interval = NEON_WARMUP_INTERVAL; // 100; we need 55 steps to get full 255 PWM = aprox 55 Seconds till the Neon tube has full brightness
          previousMillis = currentMillis;
          state = State::RUNNING;
        }
      }
      if (state == State::RUNNING) {
        uint32_t currentMillis = millis();
        if (currentMillis - previousMillis >= interval) {
          previousMillis = currentMillis;
          actual += NEON_STEP;
          //Serial.print("wa "); Serial.println(millis());
          if (actual >= NEON_MAX)
          {
            actual = NEON_MAX;
            state = State::FULL;      // lower end reached. Switch to Idle state
            //Serial.print("full "); Serial.println(millis());
          }
        }
      }
    }
};
NeonLed neon[NUMPIXELS];

// based on https://codebender.cc/sketch:271084#Neopixel%20Flames.ino
class FireLed {
  private:
    unsigned long previousMillisEffect = 0;  // timestamp for the effect
    uint16_t interval  = 10;             // how fast should the led be dimmed (=milliseconds between steps)

  public:
    int rb = 255, gb = 75, bb = 0;        // fire base colors
    int r, g, b;
    FireLed() {}

    void tick()
    {
      uint32_t currentMillis = millis();
      if (currentMillis - previousMillisEffect > interval)
      {
        previousMillisEffect = currentMillis;
        interval = random(10, 113);
        int flicker = random(0, 66);
        int brightn = random(30, 100);
        r = max(rb - flicker, 0) * brightn / 100;
        g = max(gb - flicker, 0) * brightn / 100;
        b = max(bb - flicker, 0) * brightn / 100;
        //Serial.print(r); Serial.print(" "); Serial.print(g); Serial.print(" "); Serial.println(b);
      }
    }
};
FireLed fire[NUMPIXELS];


class TvLed {
  private:
    unsigned long previousMillisEffect = 0;  // timestamp for the effect
    uint16_t interval  = 10;             // how fast should the led be dimmed (=milliseconds between steps)

  public:
    int rb = 200, gb = 190, bb = 220;        // tv base colors
    int r, g, b;
    TvLed() {}

    void tick()
    {
      uint32_t currentMillis = millis();
      if (currentMillis - previousMillisEffect > interval)
      {
        previousMillisEffect = currentMillis;
        interval = random(100, 2000);
        //CHSV(random(1, 255), random(128, 255), random(10, 255));
        int rflicker = random(0, 150);
        int gflicker = random(0, 150);
        int bflicker = random(0, 150);
        int brightn = random(30, 100);
        r = max((rb - rflicker), 0) * brightn / 100;
        g = max((gb - gflicker), 0) * brightn / 100;
        b = max((bb - bflicker), 0) * brightn / 100;
      }
    }
};
TvLed tv[NUMPIXELS];


void updateLicht(void)
{
  pixels.clear();
  for(int i=0; i<NUMPIXELS; i++)
  { // For each pixel...
    byte r=LED_COLOR_R_OFF, g=LED_COLOR_G_OFF, b=LED_COLOR_B_OFF;
    bool licht = bitRead(myData.l[SLAVE_ID], i);
    neon[i].tick();
    fire[i].tick();
    tv[i].tick();

    if (lold[i] != licht) // trigger only on change
    {
      lold[i] = licht;
      lstate[i] = 0; // reset state on switch
      ltime[i] = millis();
      if (licht)
        neon[i].on();
      else
        neon[i].off();
    }
    if (licht)
    { // switch light on
      r = lcolor[i][0];
      g = lcolor[i][1];
      b = lcolor[i][2];
      if (lmode[i] == MODE_NEON_E || lmode[i] == MODE_DEFECT_E)
      {
        int p = neon[i].actual;
        if (lmode[i] == MODE_DEFECT_E && neon[i].state == State::RUNNING)
        {
          neon[i].off();
          neon[i].on();
        }
        r = r*p/256; if (r==0 && licht) r=1;
        g = g*p/256; if (g==0 && licht) g=1;
        b = b*p/256; if (b==0 && licht) b=1;
        if (p < NEON_ON)
          licht = false; // also force pin to off
      }
      else if (lmode[i] == MODE_FIRE_E)
      {
        if (r || g || b) fire[i].rb = r;
        if (r || g || b) fire[i].gb = g;
        if (r || g || b) fire[i].bb = b;
        r = fire[i].r;
        g = fire[i].g;
        b = fire[i].b;
      }
      else if (lmode[i] == MODE_TV_E)
      {
        if (r || g || b) tv[i].rb = r;
        if (r || g || b) tv[i].gb = g;
        if (r || g || b) tv[i].bb = b;
        r = tv[i].r;
        g = tv[i].g;
        b = tv[i].b;
      }
    }
    // pixels.Color() takes RGB values, from 0,0,0 up to 255,255,255
    pixels.setPixelColor(i, pixels.Color(r, g, b));
#ifdef USE_PINS
    digitalWrite(lpin[i], (licht ? 1 : 0));
#endif
  }
  pixels.show();
}


void loadSettings()
{
  String c;
  SLAVE_ID = preferences.getUInt("slaveid", 0); //(!digitalRead(dips[2]) << 2) | (!digitalRead(dips[1]) << 1) | (!digitalRead(dips[0]));
  for (int i=0; i<NUMPIXELS; i++)
  {
    c = "l" + String(i+1, DEC) + "r";
    lcolor[i][0] = preferences.getUInt(c.c_str(), LED_COLOR_R_ON);
    c = "l" + String(i+1, DEC) + "g";
    lcolor[i][1] = preferences.getUInt(c.c_str(), LED_COLOR_G_ON);
    c = "l" + String(i+1, DEC) + "b";
    lcolor[i][2] = preferences.getUInt(c.c_str(), LED_COLOR_B_ON);
    c = "l" + String(i+1, DEC) + "m";
    lmode[i] = preferences.getUInt(c.c_str(), 0);
  }
}


#ifndef LED_BUILTIN
  #define LED_BUILTIN 2
#endif
#define ON 0
#define OFF 1


//---> COM-MODUL ESPNOW
#if (COM == COM_ESPNOW)

#ifndef ESP_OK
  #define ESP_OK 0
#endif


#ifdef ESP32
  esp_now_peer_info_t master;
  const esp_now_peer_info_t *masterNode = &master;
#else
  byte master[6];
#endif
bool (*rx_cb)(const uint8_t *, int);


void Com_init(bool (*callback)(const uint8_t *, int))
{
  rx_cb = callback;
  InitESPNow();
#ifndef ESP32
  esp_now_set_self_role(ESP_NOW_ROLE_SLAVE);
#endif
  esp_now_register_recv_cb(OnDataRecv);
}

// Init ESP Now with fallback
void InitESPNow()
{
  WiFi.disconnect();
  if (esp_now_init() == ESP_OK) {
    Serial.println("ESPNow Init Success");
  }
  else {
    Serial.println("ESPNow Init Failed");
  }
}

// callback when data is recv from Master
#ifdef ESP32
void OnDataRecv(const uint8_t *mac_addr, const uint8_t *data, int data_len) 
#else
void OnDataRecv(uint8_t * mac_addr, uint8_t *data, uint8_t data_len)
#endif
{
  if (rx_cb(data, data_len))
  {
#ifdef ESP32
    memset(&master, 0, sizeof(master));
    memcpy(master.peer_addr, mac_addr, sizeof(master.peer_addr));
    master.channel = CHANNEL;
    master.encrypt = false;
    master.ifidx = WIFI_IF_AP;
#else
    memcpy(master, mac_addr, sizeof(master));
#endif
  }
  /*
  char macStr[18];
  snprintf(macStr, sizeof(macStr), "%02x:%02x:%02x:%02x:%02x:%02x",
           mac_addr[0], mac_addr[1], mac_addr[2], mac_addr[3], mac_addr[4], mac_addr[5]);
  Serial.print("Last Packet Recv from: "); Serial.println(macStr);*/
}

void Com_loop(){}

void Com_sendFeedback(byte id)
{
#ifdef ESP32
    esp_err_t r;
    if (!esp_now_is_peer_exist(master.peer_addr))
    {
      r = esp_now_add_peer(masterNode);
      if (r != ESP_OK) Serial.println("Add master peer failed.");
    }
    const uint8_t *peer_addr = master.peer_addr;
    r = esp_now_send(peer_addr, (const uint8_t*)&id, sizeof(id));
    if (r != ESP_OK) {Serial.println(r); Serial.println("Send failed.");}
#else
    int r;
    if (!esp_now_is_peer_exist(master))
    {
      r = esp_now_add_peer((u8*)master, ESP_NOW_ROLE_CONTROLLER, CHANNEL, NULL, 0);
      if (r) Serial.println(r);
    }
    r = esp_now_send((u8*)master, (u8*)&id, sizeof(id));
    if (r) Serial.println(r);
#endif
}

#endif //<--- COM ESPNOW


//---> COM BLE
#if COM == COM_BLE

#define SERVICE_UUID        "aca83eb4-1233-485a-a8a5-b60f5be411db"
#define CHAR_MASTER_UUID    "356624bc-4be2-4e14-8ae1-f0df28b3024e"
#define MASTER_TICK_UUID    "e995501c-1091-4a55-b8f1-46749b1a42d4"
#define SLAVE1_UUID         "c4ba5132-4e95-4fd5-b363-95b857a88220"
#define SLAVE2_UUID         "19077ede-d8da-435d-9e48-561d40b6babf"
#define SLAVE3_UUID         "3d3c98cd-c3ac-4e5c-ab12-4ba616193e7a"
#define SLAVE4_UUID         "146d998e-0d6c-43ca-b42c-7294a6dd1cc4"
#define SLAVE5_UUID         "24f78ca3-d001-4dc1-96c0-ec0d134741da"
#define SLAVE6_UUID         "b069bfbb-a032-45bb-9ee5-40becbb32a5d"
#define SLAVE7_UUID         "7c93d7d7-7fff-4126-a0bc-e13a19220800"
#define SLAVE8_UUID         "63173d3f-2ce6-465c-9d6f-1c4d1a7ff740"
#define SLAVE9_UUID         "41630cda-52f5-47bb-9b46-322dc8cbd5ed"
#define SLAVE10_UUID        "ba2f57cc-d240-4420-833e-cc8040ce07cc"
static BLERemoteCharacteristic *pFromSlaves[MAX_NUM_SLAVES];
static BLERemoteCharacteristic *pFromMaster;
static byte tick, mtick;

bool (*rx_cb)(const uint8_t *, int);

// The remote service we wish to connect to.
static BLEUUID serviceUUID(SERVICE_UUID);
// The characteristic of the remote service we are interested in.
static BLEUUID    charUUID(CHAR_MASTER_UUID);

static boolean doConnect = false;
static boolean connected = false;
static boolean doScan = false;
static BLERemoteCharacteristic* pRemoteCharacteristic;
static BLEAdvertisedDevice* myDevice;

static void notifyCallback(
  BLERemoteCharacteristic* pBLERemoteCharacteristic,
  uint8_t* pData,
  size_t length,
  bool isNotify) {
    Serial.print("Notify callback for characteristic ");
    Serial.print(pBLERemoteCharacteristic->getUUID().toString().c_str());
    Serial.print(" of data length ");
    Serial.println(length);
    Serial.print("data: ");
    Serial.write(pData, length);
    Serial.println();
}

class MyClientCallback : public BLEClientCallbacks {
  void onConnect(BLEClient* pclient) {
  }

  void onDisconnect(BLEClient* pclient) {
    connected = false;
    Serial.println("onDisconnect");
  }
};

bool connectToServer() {
    Serial.print("Forming a connection to ");
    Serial.println(myDevice->getAddress().toString().c_str());
    
    BLEClient*  pClient  = BLEDevice::createClient();
    Serial.println(" - Created client");

    pClient->setClientCallbacks(new MyClientCallback());

    // Connect to the remove BLE Server.
    pClient->connect(myDevice);  // if you pass BLEAdvertisedDevice instead of address, it will be recognized type of peer device address (public or private)
    Serial.println(" - Connected to server");
    pClient->setMTU(517); //set client to request maximum MTU from server (default is 23 otherwise)
  
    // Obtain a reference to the service we are after in the remote BLE server.
    BLERemoteService* pRemoteService = pClient->getService(serviceUUID);
    if (pRemoteService == nullptr) {
      Serial.print("Failed to find our service UUID: ");
      Serial.println(serviceUUID.toString().c_str());
      pClient->disconnect();
      return false;
    }
    Serial.println(" - Found our service");


    // Obtain a reference to the characteristic in the service of the remote BLE server.
    pRemoteCharacteristic = pRemoteService->getCharacteristic(charUUID);
    if (pRemoteCharacteristic == nullptr) {
      Serial.print("Failed to find our characteristic UUID: ");
      Serial.println(charUUID.toString().c_str());
      pClient->disconnect();
      return false;
    }
    Serial.println(" - Found our characteristic");

    pFromMaster = pRemoteService->getCharacteristic((MASTER_TICK_UUID));
    pFromSlaves[0] = pRemoteService->getCharacteristic((SLAVE1_UUID));
    pFromSlaves[1] = pRemoteService->getCharacteristic((SLAVE2_UUID));
    pFromSlaves[2] = pRemoteService->getCharacteristic((SLAVE3_UUID));
    pFromSlaves[3] = pRemoteService->getCharacteristic((SLAVE4_UUID));
    pFromSlaves[4] = pRemoteService->getCharacteristic((SLAVE5_UUID));
    pFromSlaves[5] = pRemoteService->getCharacteristic((SLAVE6_UUID));
    pFromSlaves[6] = pRemoteService->getCharacteristic((SLAVE7_UUID));
    pFromSlaves[7] = pRemoteService->getCharacteristic((SLAVE8_UUID));
    pFromSlaves[8] = pRemoteService->getCharacteristic((SLAVE9_UUID));
    pFromSlaves[9] = pRemoteService->getCharacteristic((SLAVE10_UUID));

    if(pRemoteCharacteristic->canNotify())
      pRemoteCharacteristic->registerForNotify(notifyCallback);

    connected = true;
    return true;
}
/**
 * Scan for BLE servers and find the first one that advertises the service we are looking for.
 */
class MyAdvertisedDeviceCallbacks: public BLEAdvertisedDeviceCallbacks {
 /**
   * Called for each advertising BLE server.
   */
  void onResult(BLEAdvertisedDevice advertisedDevice) {
    Serial.print("BLE Advertised Device found: ");
    Serial.println(advertisedDevice.toString().c_str());

    // We have found a device, let us now see if it contains the service we are looking for.
    if (advertisedDevice.haveServiceUUID() && advertisedDevice.isAdvertisingService(serviceUUID)) {

      BLEDevice::getScan()->stop();
      myDevice = new BLEAdvertisedDevice(advertisedDevice);
      doConnect = true;
      doScan = true;
      Serial.println("found master");

    } // Found our server
  } // onResult
}; // MyAdvertisedDeviceCallbacks

void Com_init(bool (*callback)(const uint8_t *, int))
{
  rx_cb = callback;
  BLEDevice::init("");
  // Retrieve a Scanner and set the callback we want to use to be informed when we
  // have detected a new device.  Specify that we want active scanning and start the
  // scan to run for 5 seconds.
  BLEScan* pBLEScan = BLEDevice::getScan();
  pBLEScan->setAdvertisedDeviceCallbacks(new MyAdvertisedDeviceCallbacks());
  pBLEScan->setInterval(1349);
  pBLEScan->setWindow(449);
  pBLEScan->setActiveScan(true);
  pBLEScan->start(5, false);
}

void Com_loop()
{

  if (doConnect == true) {
    if (connectToServer()) {
      Serial.println("We are now connected to the BLE Server.");
    } else {
      Serial.println("We have failed to connect to the server; there is nothin more we will do.");
    }
    doConnect = false;
  }

  // If we are connected to a peer BLE Server, update the characteristic each time we are reached
  // with the current time since boot.
  if (connected) {
    // Read the value of the characteristic.
    //if(pRemoteCharacteristic->canRead()) 
    {
      std::string value = pRemoteCharacteristic->readValue();
      memcpy(&myData, value.data(), sizeof(myData));
      byte t;
      value = pFromMaster->readValue();
      memcpy(&t, value.data(), sizeof(t));
      if (t != mtick)
      {
        rx = true;
        mtick = t;
        toggle = !toggle;
      }
    }
    pFromSlaves[SLAVE_ID]->writeValue(&tick, sizeof(tick));
  }
  else if(doScan) {
    BLEDevice::getScan()->start(0);  // this is just example to start scan after disconnect, most likely there is better way to do it in arduino
  }
}

void Com_sendFeedback(byte id)
{
  if (connected) {
    tick++;
    pFromSlaves[id]->writeValue(&tick, sizeof(tick));
  }
}
#endif //<--- COM BLE


String buildMainform()
{
  String c, buf = mainForm;
  buf.replace("#VERSION", VERSION);
  buf.replace("#SLAVEID", String(SLAVE_ID, DEC));
  for (int i=0; i<NUMPIXELS; i++)
  {
    c = "#CL" + String(i+1, DEC) + "X";
    if (myData.l[SLAVE_ID] & (1<<i))
      buf.replace(c.c_str(), "/aus"+String((i+1), DEC)+">ausschalten");
    else
      buf.replace(c.c_str(), "/an"+String((i+1), DEC)+">anschalten");
    c = "#L" + String(i+1, DEC) + "C";
    if (myData.l[SLAVE_ID] & (1<<i))
      buf.replace(c.c_str(), "checked");
    c = "#L" + String(i+1, DEC) + "R";
    buf.replace(c.c_str(), String(lcolor[i][0], DEC));
    c = "#L" + String(i+1, DEC) + "G";
    buf.replace(c.c_str(), String(lcolor[i][1], DEC));
    c = "#L" + String(i+1, DEC) + "B";
    buf.replace(c.c_str(), String(lcolor[i][2], DEC));
    c = "#L" + String(i+1, DEC) + "M";
    buf.replace(c.c_str(), String(lmode[i], DEC));
  }
  return buf;
}


// config AP SSID
void configDeviceAP(byte id) {
  char id_s[2] = {0,0};
  id_s[0] = '0'+id;
  String Prefix = "Slave:";
  String SlaveID = id_s;
  String Mac = WiFi.macAddress();
  String SSID = "Slave" + SlaveID + ":" + Mac;
  String Password = WIFI_PASSWORD;
  bool result = WiFi.softAP(SSID.c_str(), Password.c_str(), CHANNEL, 0);
  if (!result) {
    Serial.println("AP Config failed.");
  } else {
    Serial.println("AP Config Success. Broadcasting with AP: " + String(SSID));
  }
  IPAddress IP = IPAddress (10, 0, 0, 1);
  IPAddress NMask = IPAddress (255, 255, 255, 0);
  WiFi.softAPConfig(IP, IP, NMask);
  IPAddress myIP = WiFi.softAPIP();
  Serial.print("AP IP address: ");
  Serial.println(myIP);

  server.on("/", HTTP_GET, []() {
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/an1", HTTP_GET, []() {
    myData.l[SLAVE_ID] |= 0x01U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/aus1", HTTP_GET, []() {
    myData.l[SLAVE_ID] &= ~0x01U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/an2", HTTP_GET, []() {
    myData.l[SLAVE_ID] |= 0x02U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/aus2", HTTP_GET, []() {
    myData.l[SLAVE_ID] &= ~0x02U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/an3", HTTP_GET, []() {
    myData.l[SLAVE_ID] |= 0x04U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/aus3", HTTP_GET, []() {
    myData.l[SLAVE_ID] &= ~0x04U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/an4", HTTP_GET, []() {
    myData.l[SLAVE_ID] |= 0x08U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/aus4", HTTP_GET, []() {
    myData.l[SLAVE_ID] &= ~0x08U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/an5", HTTP_GET, []() {
    myData.l[SLAVE_ID] |= 0x010U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/aus5", HTTP_GET, []() {
    myData.l[SLAVE_ID] &= ~0x010U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/an6", HTTP_GET, []() {
    myData.l[SLAVE_ID] |= 0x020U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/aus6", HTTP_GET, []() {
    myData.l[SLAVE_ID] &= ~0x020U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/an7", HTTP_GET, []() {
    myData.l[SLAVE_ID] |= 0x040U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/aus7", HTTP_GET, []() {
    myData.l[SLAVE_ID] &= ~0x040U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/an8", HTTP_GET, []() {
    myData.l[SLAVE_ID] |= 0x080U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/aus8", HTTP_GET, []() {
    myData.l[SLAVE_ID] &= ~0x080U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/an9", HTTP_GET, []() {
    myData.l[SLAVE_ID] |= 0x100U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/aus9", HTTP_GET, []() {
    myData.l[SLAVE_ID] &= ~0x100U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/an10", HTTP_GET, []() {
    myData.l[SLAVE_ID] |= 0x200U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });
  server.on("/aus10", HTTP_GET, []() {
    myData.l[SLAVE_ID] &= ~0x200U;
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", buildMainform());
  });

  server.on("/get", HTTP_GET, []() {
    byte sid = server.arg("slaveid").toInt(); preferences.putUInt("slaveid", sid);
    String c;
    for (int i=0; i<NUMPIXELS; i++)
    {
      c = "l" + String(i+1, DEC);
      if (server.arg(c.c_str()) == "on")
        myData.l[SLAVE_ID] |= 1<<i;
      else
        myData.l[SLAVE_ID] &= ~(1<<i);
      c = "l" + String(i+1, DEC) + "r";
      lcolor[i][0] = server.arg(c.c_str()).toInt(); preferences.putUInt(c.c_str(), lcolor[i][0]);
      c = "l" + String(i+1, DEC) + "g";
      lcolor[i][1] = server.arg(c.c_str()).toInt(); preferences.putUInt(c.c_str(), lcolor[i][1]);
      c = "l" + String(i+1, DEC) + "b";
      lcolor[i][2] = server.arg(c.c_str()).toInt(); preferences.putUInt(c.c_str(), lcolor[i][2]);
      c = "l" + String(i+1, DEC) + "m";
      lmode[i] = server.arg(c.c_str()).toInt(); preferences.putUInt(c.c_str(), lmode[i]);
    }
  
    server.sendHeader("Connection", "close");
    if (sid != SLAVE_ID)
    {
      server.send(200, "text/html", reboot);
      ESP.restart();
    }
    else
    {
      server.send(200, "text/html", noreboot);
    }
  });

  server.on("/prepupdate", HTTP_GET, []() {
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", prepupdateForm);
    //esp_now_deinit();
    updatemode = true;
  });

  ESP2SOTA.begin(&server);
  server.begin();
}


bool rx_callback_fn(const uint8_t *data, int data_len)
{
  bool res = false;
  if (data_len == sizeof(myData))
  {
    res = true;
    memcpy(&myData, data, sizeof(myData));
    toggle = !toggle;
    rx = true;
  }
  return res;
}


void setup()
{
  preferences.begin("lichtempf", false);
  Serial.begin(115200);
  pinMode(LED_BUILTIN, OUTPUT);
#ifdef USE_PINS
  for (int i=0; i<sizeof(lpin); i++)
  {
    pinMode(lpin[i], OUTPUT);
    digitalWrite(lpin[i], 0);
  }
#endif
  pixels.begin();
  WiFi.mode(WIFI_AP);
  loadSettings();
  Serial.print("Slave-ID: "); Serial.println(SLAVE_ID);
  configDeviceAP(SLAVE_ID);
  Serial.print("AP MAC: "); Serial.println(WiFi.softAPmacAddress());
  Com_init(&rx_callback_fn);
  digitalWrite(LED_BUILTIN, ON); // show init is ok
}


void loop()
{
  digitalWrite(LED_BUILTIN, toggle ? ON : OFF); // show slave is alive and receiving
  if (!updatemode)
  {
    Com_loop();
    updateLicht();
    if (rx)
    {
      rx = false;
      Com_sendFeedback(SLAVE_ID);
    }
  }
  server.handleClient();
}
