/**Lichtsender V2

Uses:
   ESPNOW - Basic communication - Master
   Date: 26th September 2017
   Author: Arvind Ravulavaru <https://github.com/arvindr21>
   Purpose: ESPNow Communication between a Master ESP32 and multiple ESP32 Slaves
   Description: This sketch consists of the code for the Master module.
   Resources: (A bit outdated)
   a. https://espressif.com/sites/default/files/documentation/esp-now_user_guide_en.pdf
   b. http://www.esploradores.com/practica-6-conexion-esp-now/

   << This Device Master >>

   Flow: Master
   Step 1 : ESPNow Init on Master and set it in STA mode
   Step 2 : Start scanning for Slave ESP32 (we have added a prefix of `slave` to the SSID of slave for an easy setup)
   Step 3 : Once found, add Slave as peer
   Step 4 : Register for send callback
   Step 5 : Start Transmitting data from Master to Slave(s)

   Flow: Slave
   Step 1 : ESPNow Init on Slave
   Step 2 : Update the SSID of Slave with a prefix of `slave`
   Step 3 : Set Slave in AP mode
   Step 4 : Register for receive callback and wait for data
   Step 5 : Once data arrives, print it in the serial monitor

   Note: Master and Slave have been defined to easily understand the setup.
         Based on the ESPNOW API, there is no concept of Master and Slave.
         Any devices can act as master or salve.
*/
#define COM_ESPNOW 0
#define COM_BLE 1

#define COM COM_ESPNOW


#if COM == COM_ESPNOW
  #include <esp_now.h>
#endif
#if COM == COM_BLE
  #include <BLEDevice.h>
  #include <BLEUtils.h>
  #include <BLEServer.h>
#endif
#include <WiFi.h>
#include <Adafruit_NeoPixel.h> // added pixels to show selected slave (house)
#include <ESP2SOTA.h>


const char* ssid = "LICHTSENDER";
const char* password = "esp32licht";
String mainForm = 
"<html><head><title>Updater</title></head><body><h3>Selfupdate Lichtsender</h3><p>"
"<a href=/update>Update Firmware</a><br><br><a href=/cancel>Abbrechen und Neustarten</a></body></html>";
WebServer server(80);
bool updatemode;


#define SERIAL_VERBOSE 0 // show debug print
#define LED_BUILTIN 2 // onboard led

#define MAX_NUM_SLAVES 10 // max support of # slaves
#define NUMBUTTONS 11 // available buttons, first is slave select
#define LONGWAIT 20 // for long button press ca. 2sec
#define LOOPWAIT 80 // ms to wait in loop
#define SLAVE_FW_RESET_TIMEOUT 10
#define COLOR_INACTIVE {0,0,1}
#define COLOR_INACTIVE_DETECTED {0,1,0}
#define COLOR_ACTIVE {0,0,100}
#define COLOR_ACTIVE_DETECTED {0,40,0}
#define COLOR_LIGHT_ON {100,80,30}
#define COLOR_LIGHT_OFF {0,0,0}

#define PIN       13 // neopixel pin
#define SLAVESELPIXELS 4 // smartleds stripes are in series
#define LIGHTSTATEPIXELS 10
#define NUMPIXELS (SLAVESELPIXELS+LIGHTSTATEPIXELS) // to show the selected slave - also the max number of controllable slaves

Adafruit_NeoPixel pixels(NUMPIXELS, PIN, NEO_GRB + NEO_KHZ800);


const byte btnpins[NUMBUTTONS] = {
  22, // slave select (long press for rescan slaves)
  27, // licht 1-10 (long press for fw update on slave id 0-9)
  14,
  12,
  23,
  32,
  33,
  25,
  26,
  19,
  18
};


typedef struct licht_st
{
  unsigned short l[MAX_NUM_SLAVES];
} licht_s;


#if COM == COM_ESPNOW
//---> COM-MODUL ESPNOW
#define CHANNEL 1

esp_now_peer_info_t slaves[MAX_NUM_SLAVES];
int SlaveCnt;
bool rescan_;
byte slave_timer[MAX_NUM_SLAVES];


// callback when data is sent from Master to Slave
void OnDataSent(const uint8_t *mac_addr, esp_now_send_status_t status)
{
  char macStr[18];
  if (SERIAL_VERBOSE) {
    snprintf(macStr, sizeof(macStr), "%02x:%02x:%02x:%02x:%02x:%02x",
            mac_addr[0], mac_addr[1], mac_addr[2], mac_addr[3], mac_addr[4], mac_addr[5]);
    Serial.print("Last Packet Sent to: "); Serial.println(macStr);
    Serial.print("Last Packet Send Status: "); Serial.println(status == ESP_NOW_SEND_SUCCESS ? "Delivery Success" : "Delivery Fail");
  }
}

// receive feedback
void OnDataRecv(const uint8_t *mac_addr, const uint8_t *data, int data_len) 
{
  if (data_len == sizeof(byte))
  {
    byte slaveid = data[0];
    if (slaveid < MAX_NUM_SLAVES)
      slave_timer[slaveid] = 10;
  }
  if (SERIAL_VERBOSE) {
    char macStr[18];
    snprintf(macStr, sizeof(macStr), "%02x:%02x:%02x:%02x:%02x:%02x",
            mac_addr[0], mac_addr[1], mac_addr[2], mac_addr[3], mac_addr[4], mac_addr[5]);
    Serial.print("Feedback Recv from: "); Serial.print(macStr); Serial.print("; ID = ");
    Serial.println(data[0]);
  }
}

static void Com_init()
{
  SlaveCnt = 0;
  // static CommEspNow* obj = this;
  InitESPNow();
  // Once ESPNow is successfully Init, we will register for Send CB to
  // get the status of Trasnmitted packet
  esp_now_register_send_cb(OnDataSent);
  esp_now_register_recv_cb(OnDataRecv);
  ScanForSlave(); // scan once on startup
  if (SlaveCnt > 0) { // check if slave channel is defined
    // `slave` is defined
    // Add slave as peer if it has not been added already
    manageSlave();
    // pair success or already paired
  }
}

static void Com_send(const uint8_t* data, int len)
{
  // If Slave is found, it would be populate in `slave` variable
  // We will check if `slave` is defined and then we proceed further
  if (SlaveCnt > 0)
  { // check if slave channel is defined
    // `slave` is defined
    // Add slave as peer if it has not been added already
    manageSlave();
    // pair success or already paired
    // Send data to device
    sendData(data, len);
  } 
  if (SlaveCnt == 0 || rescan_)
  {
    // In the loop we scan for slave
    ScanForSlave();
    rescan_ = 0;
  }
}

static void Com_rescan()
{
  rescan_ = true;
}

static bool Com_checkSlave(byte num)
{
  if (num >= MAX_NUM_SLAVES)
    num = MAX_NUM_SLAVES-1;
  return (slave_timer[num] > 0);
}

// Init ESP Now with fallback
static void InitESPNow()
{
  WiFi.disconnect();
  if (esp_now_init() == ESP_OK) {
    if (SERIAL_VERBOSE) Serial.println("ESPNow Init Success");
  }
  else {
    if (SERIAL_VERBOSE) Serial.println("ESPNow Init Failed");
  }
}

// Scan for slaves in AP mode
static void ScanForSlave()
{
  int8_t scanResults = WiFi.scanNetworks();
  //reset slaves
  memset(slaves, 0, sizeof(slaves));
  SlaveCnt = 0;
  if (SERIAL_VERBOSE) Serial.println("");
  if (scanResults == 0) {
    if (SERIAL_VERBOSE) Serial.println("No WiFi devices in AP Mode found");
  } else {
    if (SERIAL_VERBOSE) {Serial.print("Found "); Serial.print(scanResults); Serial.println(" devices ");}
    for (int i = 0; i < scanResults; ++i) {
      // Print SSID and RSSI for each device found
      String SSID = WiFi.SSID(i);
      int32_t RSSI = WiFi.RSSI(i);
      String BSSIDstr = WiFi.BSSIDstr(i);

      if (SERIAL_VERBOSE) {Serial.print(i + 1); Serial.print(": "); Serial.print(SSID); Serial.print(" ["); Serial.print(BSSIDstr); Serial.print("]"); Serial.print(" ("); Serial.print(RSSI); Serial.print(")"); Serial.println("");}
      delay(10);
      // Check if the current device starts with `Slave`
      if (SSID.indexOf("Slave") == 0) {
        // SSID of interest
        byte slaveid = SSID.c_str()[5] - '0';
        Serial.print(SSID);
        Serial.println(slaveid);
        if (SERIAL_VERBOSE) {Serial.print(i + 1); Serial.print(": "); Serial.print(SSID); Serial.print(" ["); Serial.print(BSSIDstr); Serial.print("]"); Serial.print(" ("); Serial.print(RSSI); Serial.print(")"); Serial.println("");}
        // Get BSSID => Mac Address of the Slave
        int mac[6];

        if ( 6 == sscanf(BSSIDstr.c_str(), "%x:%x:%x:%x:%x:%x",  &mac[0], &mac[1], &mac[2], &mac[3], &mac[4], &mac[5] ) ) {
          for (int ii = 0; ii < 6; ++ii ) {
            slaves[SlaveCnt].peer_addr[ii] = (uint8_t) mac[ii];
          }
        }
        slaves[SlaveCnt].channel = CHANNEL; // pick a channel
        slaves[SlaveCnt].encrypt = 0; // no encryption
        SlaveCnt++;
      }
    }
  }

  if (SlaveCnt > 0) {
    if (SERIAL_VERBOSE) Serial.print(SlaveCnt); 
    if (SERIAL_VERBOSE) Serial.println(" Slave(s) found, processing..");
  } else {
    if (SERIAL_VERBOSE) Serial.println("No Slave Found, trying again.");
  }

  // clean up ram
  WiFi.scanDelete();
}

// Check if the slave is already paired with the master.
// If not, pair the slave with master
static void manageSlave()
{
  if (SlaveCnt > 0) {
    for (int i = 0; i < SlaveCnt; i++) {
      if (SERIAL_VERBOSE) {
        Serial.print("Processing: ");
        for (int ii = 0; ii < 6; ++ii ) {
          Serial.print((uint8_t) slaves[i].peer_addr[ii], HEX);
          if (ii != 5) Serial.print(":");
        }
        Serial.print(" Status: ");
      }
      // check if the peer exists
      bool exists = esp_now_is_peer_exist(slaves[i].peer_addr);
      if (exists) {
        // Slave already paired.
        if (SERIAL_VERBOSE) Serial.println("Already Paired");
      } else {
        // Slave not paired, attempt pair
        esp_err_t addStatus = esp_now_add_peer(&slaves[i]);
        if (addStatus == ESP_OK) {
          // Pair success
          if (SERIAL_VERBOSE) Serial.println("Pair success");
        } else if (addStatus == ESP_ERR_ESPNOW_NOT_INIT) {
          // How did we get so far!!
          if (SERIAL_VERBOSE) Serial.println("ESPNOW Not Init");
        } else if (addStatus == ESP_ERR_ESPNOW_ARG) {
          if (SERIAL_VERBOSE) Serial.println("Add Peer - Invalid Argument");
        } else if (addStatus == ESP_ERR_ESPNOW_FULL) {
          if (SERIAL_VERBOSE) Serial.println("Peer list full");
        } else if (addStatus == ESP_ERR_ESPNOW_NO_MEM) {
          if (SERIAL_VERBOSE) Serial.println("Out of memory");
        } else if (addStatus == ESP_ERR_ESPNOW_EXIST) {
          if (SERIAL_VERBOSE) Serial.println("Peer Exists");
        } else {
          if (SERIAL_VERBOSE) Serial.println("interface error (check ap mode, idifx)");
        }
        //delay(100);
      }
    }
  } else {
    // No slave found to process
    if (SERIAL_VERBOSE) Serial.println("No Slave found to process");
  }
}

// send data
static void sendData(const uint8_t* data, int len) 
{
  esp_err_t result = esp_now_send(0, data, len); // peer_addr set to broadcast
  if (SERIAL_VERBOSE) Serial.print("Send Status: ");
  if (result == ESP_OK) {
    if (SERIAL_VERBOSE) Serial.println("Success");
  }
  else {
    if (SERIAL_VERBOSE) Serial.println("Fail");
  }
  for (int i=0; i<MAX_NUM_SLAVES; i++)
  {
    if (slave_timer[i] > 0)
      slave_timer[i]--;
  }
}
#endif //<--- END COM-MODUL ESPNOW


//---> COM BLE
#if COM == COM_BLE

//#define SERVICE_UUID        "4fafc201-1fb5-459e-8fcc-c5c9c331914b"
//#define CHARACTERISTIC_UUID "beb5483e-36e1-4688-b7f5-ea07361b26a8"
#define SERVICE_UUID        "aca83eb4-1233-485a-a8a5-b60f5be411db"
#define CHARACTERISTIC_UUID "356624bc-4be2-4e14-8ae1-f0df28b3024e"
#define MASTER_TICK_UUID    "e995501c-1091-4a55-b8f1-46749b1a42d4"
#define SLAVE1_UUID         "c4ba5132-4e95-4fd5-b363-95b857a88220"
#define SLAVE2_UUID         "19077ede-d8da-435d-9e48-561d40b6babf"
#define SLAVE3_UUID         "3d3c98cd-c3ac-4e5c-ab12-4ba616193e7a"
#define SLAVE4_UUID         "146d998e-0d6c-43ca-b42c-7294a6dd1cc4"
#define SLAVE5_UUID         "24f78ca3-d001-4dc1-96c0-ec0d134741da"
#define SLAVE6_UUID         "b069bfbb-a032-45bb-9ee5-40becbb32a5d"
#define SLAVE7_UUID         "7c93d7d7-7fff-4126-a0bc-e13a19220800"
#define SLAVE8_UUID         "63173d3f-2ce6-465c-9d6f-1c4d1a7ff740"
#define SLAVE9_UUID         "41630cda-52f5-47bb-9b46-322dc8cbd5ed"
#define SLAVE10_UUID        "ba2f57cc-d240-4420-833e-cc8040ce07cc"
#define SLAVE_TIMEOUT 10

BLECharacteristic *pToSlaves;
BLECharacteristic *pMasterTick;
BLECharacteristic *pFromSlaves[MAX_NUM_SLAVES];
byte slave_tick[MAX_NUM_SLAVES], mtick;
int slave_timeout[MAX_NUM_SLAVES];


void Com_init()
{
  BLEDevice::init("Lichtsender");
  BLEServer *pServer = BLEDevice::createServer();
  BLEService *pService = pServer->createService(SERVICE_UUID);
  pToSlaves = pService->createCharacteristic(
                                         CHARACTERISTIC_UUID,
                                         BLECharacteristic::PROPERTY_READ |
                                         BLECharacteristic::PROPERTY_WRITE
                                       );
  
  pMasterTick = pService->createCharacteristic(
                                         MASTER_TICK_UUID,
                                         BLECharacteristic::PROPERTY_READ |
                                         BLECharacteristic::PROPERTY_WRITE
                                       );
  pFromSlaves[0] = pService->createCharacteristic(SLAVE1_UUID, BLECharacteristic::PROPERTY_READ | BLECharacteristic::PROPERTY_WRITE);
  pFromSlaves[1] = pService->createCharacteristic(SLAVE2_UUID, BLECharacteristic::PROPERTY_READ | BLECharacteristic::PROPERTY_WRITE);
  pFromSlaves[2] = pService->createCharacteristic(SLAVE3_UUID, BLECharacteristic::PROPERTY_READ | BLECharacteristic::PROPERTY_WRITE);
  pFromSlaves[3] = pService->createCharacteristic(SLAVE4_UUID, BLECharacteristic::PROPERTY_READ | BLECharacteristic::PROPERTY_WRITE);
  pFromSlaves[4] = pService->createCharacteristic(SLAVE5_UUID, BLECharacteristic::PROPERTY_READ | BLECharacteristic::PROPERTY_WRITE);
  pFromSlaves[5] = pService->createCharacteristic(SLAVE6_UUID, BLECharacteristic::PROPERTY_READ | BLECharacteristic::PROPERTY_WRITE);
  pFromSlaves[6] = pService->createCharacteristic(SLAVE7_UUID, BLECharacteristic::PROPERTY_READ | BLECharacteristic::PROPERTY_WRITE);
  pFromSlaves[7] = pService->createCharacteristic(SLAVE8_UUID, BLECharacteristic::PROPERTY_READ | BLECharacteristic::PROPERTY_WRITE);
  pFromSlaves[8] = pService->createCharacteristic(SLAVE9_UUID, BLECharacteristic::PROPERTY_READ | BLECharacteristic::PROPERTY_WRITE);
  pFromSlaves[9] = pService->createCharacteristic(SLAVE10_UUID, BLECharacteristic::PROPERTY_READ | BLECharacteristic::PROPERTY_WRITE);

  //pToSlaves->setValue(0);
  pService->start();
  // BLEAdvertising *pAdvertising = pServer->getAdvertising();  // this still is working for backward compatibility
  BLEAdvertising *pAdvertising = BLEDevice::getAdvertising();
  pAdvertising->addServiceUUID(SERVICE_UUID);
  pAdvertising->setScanResponse(true);
  pAdvertising->setMinPreferred(0x06);  // functions that help with iPhone connections issue
  pAdvertising->setMinPreferred(0x12);
  BLEDevice::startAdvertising();
}

static void Com_send(const uint8_t* data, int len)
{
  pToSlaves->setValue((uint8_t*)data, len);
  mtick++;
  pMasterTick->setValue((uint8_t*)&mtick, sizeof(mtick));
}

static void Com_rescan()
{
  BLEDevice::stopAdvertising();
  delay(1000);
  BLEDevice::startAdvertising();
}

static bool Com_checkSlave(byte num)
{
  byte tick;
  std::string value = pFromSlaves[num]->getValue();
  memcpy(&tick, value.data(), sizeof(tick));
  if (tick != slave_tick[num])
  {
    slave_tick[num] = tick;
    slave_timeout[num] = SLAVE_TIMEOUT;
  }
  else if (slave_timeout[num] > 0)
    slave_timeout[num]--;
  return slave_timeout[num] > 0;
}
#endif //<---- END COM BLE


licht_s myData;
byte selected_slave;
int longpress[NUMBUTTONS];
bool buttons[NUMBUTTONS], old_btn[NUMBUTTONS];


void checkButtons(void)
{
  // check button states
  for (int i=0; i<NUMBUTTONS; i++)
  {
    buttons[i] = !digitalRead(btnpins[i]);
    if (buttons[i]) // pressed button counts longpress time
    {
      if (++longpress[i] >= LONGWAIT)
      {
        if (i == 0) // button 0 triggers slave scan
        {
          Com_rescan();
        }
      }
    }
    else
    {
      longpress[i] = 0; // reset longpress time when button is released
    }

    if (old_btn[i] != buttons[i]) // react on button state change
    {
      old_btn[i] = buttons[i];
      if (buttons[i]) // button i pressed
      {
        // toggle light in myData
        if (i > 0)
        {
          byte lnum = i-1; // button 0 is not a light toggler
          bitWrite(myData.l[selected_slave], lnum, !bitRead(myData.l[selected_slave], lnum));
        }
        else // first button selects slave and long press scans for new slaves
        {
          selected_slave++;
          if (selected_slave == SLAVESELPIXELS)
            selected_slave = 0;
        }
      }
    }
  }
}


void updatePixel(void)
{
  const byte c_inactive[] = COLOR_INACTIVE;
  const byte c_inactive_d[] = COLOR_INACTIVE_DETECTED;
  const byte c_active[] = COLOR_ACTIVE;
  const byte c_active_d[] = COLOR_ACTIVE_DETECTED;
  byte c[3];
  pixels.clear();
  for (int i=0; i<SLAVESELPIXELS; i++)
  { // For each slave select pixel...
    memcpy(c, c_inactive, sizeof(c));
    if (Com_checkSlave(i))
      memcpy(c, c_inactive_d, sizeof(c));
    if (i == selected_slave)
    {
      if (Com_checkSlave(i))
        memcpy(c, c_active_d, sizeof(c));
      else
        memcpy(c, c_active, sizeof(c));
    }
    // pixels.Color() takes RGB values, from 0,0,0 up to 255,255,255
    pixels.setPixelColor(i, pixels.Color(c[0],c[1],c[2]));
  }
  const byte c_on[] = COLOR_LIGHT_ON;
  const byte c_off[] = COLOR_LIGHT_OFF;
  int k = 0;
  for (int i=SLAVESELPIXELS; i<NUMPIXELS; i++)
  { // For each light state pixel...
    memcpy(c, c_off, sizeof(c));
    if (bitRead(myData.l[selected_slave], k))
      memcpy(c, c_on, sizeof(c));
    k++;
    // pixels.Color() takes RGB values, from 0,0,0 up to 255,255,255
    pixels.setPixelColor(i, pixels.Color(c[0],c[1],c[2]));
  }
  pixels.show();
}


void setup()
{
  pinMode(LED_BUILTIN, OUTPUT);
  for (int i=0; i<sizeof(btnpins); i++)
  {
    pinMode(btnpins[i], INPUT_PULLUP);
  }
  Serial.begin(115200);
  pixels.begin();
  delay(100);
  WiFi.mode(WIFI_AP);
  WiFi.softAP(ssid, password);
  IPAddress IP = IPAddress (10, 0, 0, 1);
  IPAddress NMask = IPAddress (255, 255, 255, 0);
  WiFi.softAPConfig(IP, IP, NMask);
  IPAddress myIP = WiFi.softAPIP();
  Serial.print("AP IP address: ");
  Serial.println(myIP);

  server.on("/", HTTP_GET, []() {
    server.sendHeader("Connection", "close");
    server.send(200, "text/html", mainForm);
    updatemode = true; // on server access switch to update mode
  });
  server.on("/cancel", HTTP_GET, []() {
    server.sendHeader("Connection", "close");
    server.send(200, "text/plain", "..Neustart..");
    ESP.restart();
  });

  ESP2SOTA.begin(&server);
  server.begin();
  Com_init();
}


bool toggle;
void loop() 
{
  toggle = !toggle;
  digitalWrite(LED_BUILTIN, toggle);  // Toggle the LED
  server.handleClient();
  if (updatemode)
  {
    delay(5);
  }
  else
  {
    checkButtons();
    updatePixel();
    Com_send((const uint8_t *)&myData, sizeof(myData));
    delay(LOOPWAIT);
  }
}
